# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Basic A-RGB Color class."""

from copy import copy

try:
    from System import Drawing  # type: ignore
except ImportError:
    pass  # outside .NET

try:
    from typing import Dict, Optional
except ImportError:
    pass  # Python 2.7


class PhColor(object):
    def __init__(self):
        self.a = 0
        self.r = 0
        self.g = 0
        self.b = 0

    @classmethod
    def from_argb(cls, a, r, g, b):
        # type: (int, int, int, int) -> PhColor
        new_color = cls()
        new_color.a = int(max(0, min(a, 255)))
        new_color.r = int(max(0, min(r, 255)))
        new_color.g = int(max(0, min(g, 255)))
        new_color.b = int(max(0, min(b, 255)))
        return new_color

    @classmethod
    def from_rgb(cls, r, g, b):
        # type: (int, int, int) -> PhColor
        return cls.from_argb(255, r, g, b)

    @classmethod
    def from_system_color(cls, color):
        # type: (Drawing.Color) -> PhColor
        new_color = cls()
        new_color.a = int(max(0, min(color.A, 255)))
        new_color.r = int(max(0, min(color.R, 255)))
        new_color.g = int(max(0, min(color.G, 255)))
        new_color.b = int(max(0, min(color.B, 255)))
        return new_color

    def to_dict(self):
        # type: () -> Dict[str, float]
        return {
            "a": self.a,
            "r": self.r,
            "g": self.g,
            "b": self.b,
        }

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Optional[Dict[str, float]]) -> Optional[PhColor]
        if not _input_dict:
            return None

        new_color = cls()
        new_color.a = int(_input_dict["a"])
        new_color.r = int(_input_dict["r"])
        new_color.g = int(_input_dict["g"])
        new_color.b = int(_input_dict["b"])

        return new_color

    def __repr__(self):
        # type: () -> str
        return str(self)

    def __str__(self):
        # type: () -> str
        return "Color(a={}, r={}, g={}, b={})".format(self.a, self.r, self.g, self.b)

    def __eq__(self, other):
        # type: (PhColor) -> bool
        return self.a == other.a and self.r == other.r and self.g == other.g and self.b == other.b

    def ToString(self):
        # type: () -> str
        return str(self)

    def __copy__(self):
        # type: () -> PhColor
        new_obj = PhColor()
        new_obj.a = copy(self.a)
        new_obj.r = copy(self.r)
        new_obj.g = copy(self.g)
        new_obj.b = copy(self.b)
        return new_obj

    def duplicate(self):
        # type: () -> PhColor
        return self.__copy__()



"""Extra Boundary Condition objects for Passive House models.

Note to developers:
    See _extend_honeybee_energy_ph to see where these boundary conditions are added to
    honeybee.boundarycondition module.
"""

try:
    from typing import Any
except ImportError:
    pass  # IronPython2.7

try:
    from honeybee.altnumber import Autocalculate, autocalculate
except ImportError as e:
    raise ImportError("\nFailed to import honeybee:\n\t{}".format(e))

try:
    from honeybee_energy.boundarycondition import OtherSideTemperature
except ImportError as e:
    raise ImportError("\nFailed to import honeybee-energy:\n\t{}".format(e))


class PhAdditionalZone(OtherSideTemperature):
    """For surfaces exposed to attached zones with different temperature settings.

    This class extends the basic honeybee-energy OtherSideTemperature boundary condition
    by adding properties specific to Passive House modeling, such as zone-name, zone-type,
    and temperature-reduction-factor.
    """

    __slots__ = ("identifier", "zone_name", "zone_type", "zone_id_num", "temperature_reduction_factor")

    def __init__(
        self,
        identifier="",
        temperature=autocalculate,
        heat_transfer_coefficient=0,
        zone_name="",
        zone_type="",
        temperature_reduction_factor=1.0,
    ):
        # type: (str, float | Autocalculate, float, str, str, float) -> None
        """Initialize PhAdditionalZone boundary condition."""
        super(PhAdditionalZone, self).__init__(
            temperature=temperature, heat_transfer_coefficient=heat_transfer_coefficient
        )
        self.identifier = identifier
        self.zone_id_num = 0
        self.zone_name = zone_name
        self.zone_type = zone_type
        self.temperature_reduction_factor = temperature_reduction_factor

    def to_dict(self):
        # type: () -> dict
        parent_dict = super(PhAdditionalZone, self).to_dict()  # type: dict[str, Any]
        parent_dict.update(
            {
                "identifier": self.identifier,
                "zone_name": self.zone_name,
                "zone_type": self.zone_type,
                "zone_id_num": self.zone_id_num,
                "temperature_reduction_factor": self.temperature_reduction_factor,
            }
        )
        return parent_dict

    @classmethod
    def from_dict(cls, data):
        # type: (dict) -> PhAdditionalZone
        """Initialize PhAdditionalZone BoundaryCondition from a dictionary.

        Args:
            data: A dictionary representation of the boundary condition.
        """
        assert (
            data["type"] == "PhAdditionalZone"
        ), "Expected dictionary for PhAdditionalZone " "boundary condition. Got {}.".format(data["type"])
        temperature = (
            autocalculate
            if "temperature" not in data or data["temperature"] == autocalculate.to_dict()
            else data["temperature"]
        )
        return cls(
            identifier=data["identifier"],
            temperature=temperature,
            heat_transfer_coefficient=data["heat_transfer_coefficient"],
            zone_name=data["zone_name"],
            zone_type=data["zone_type"],
            temperature_reduction_factor=data["temperature_reduction_factor"],
        )

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        temperature_key = "Autocalculate" if isinstance(self.temperature, Autocalculate) else self.temperature
        return (
            temperature_key,
            self.heat_transfer_coefficient,
            self.zone_name,
            self.zone_type,
            self.temperature_reduction_factor,
            self.zone_id_num,
            self.identifier,
        )

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, PhAdditionalZone) and self.__key() == other.__key()



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""HB-PH Model Properties."""

try:
    from typing import Any, Dict, Tuple
except ImportError:
    pass  # Python 2.7

try:
    from itertools import izip as zip  # type: ignore
except ImportError:
    pass  # Python3

try:
    from honeybee import extensionutil
except ImportError as e:
    raise ImportError("\nFailed to import honeybee:\n\t{}".format(e))

try:
    from honeybee import extensionutil
except ImportError as e:
    raise ImportError("\nFailed to import honeybee:\n\t{}".format(e))

try:
    from honeybee_ph.bldg_segment import BldgSegment
    from honeybee_ph.team import ProjectTeam
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_ph:\n\t{}".format(e))


class ModelPhProperties(object):
    def __init__(self, _host):
        self._host = _host
        self.id_num = 0
        self.team = ProjectTeam()

    @property
    def host(self):
        return self._host

    def duplicate(self, new_host=None):
        # type: (Any) -> ModelPhProperties
        _host = new_host or self._host
        new_properties_obj = ModelPhProperties(_host)
        new_properties_obj.id_num = self.id_num
        new_properties_obj.team = self.team.duplicate()

        return new_properties_obj

    def __str__(self):
        return "Model Passive House Properties: [host: {}]".format(self.host.display_name)

    def __repr__(self):
        return str(self)

    def ToString(self):
        return self.__repr__()

    def _get_bldg_segment_dicts(self):
        # type: () -> list[dict[str, Any]]
        """Return a list of all the bldg_segments found on the model's rooms as dicts.
            This is used when writing an 'Abridged' HBJSON file.

        Arguments:
        ----------
            * None

        Returns:
        --------
            * list[dict[str, Any]]: A list of all the bldg_segments found on the model's
                rooms as dicts.
        """
        # -- Collect all the unique BldgSegments in the Model's Rooms
        ph_bldg_segments = {
            rm.properties.ph.ph_bldg_segment.identifier: rm.properties.ph.ph_bldg_segment for rm in self.host.rooms
        }
        return [seg.to_dict() for seg in ph_bldg_segments.values()]

    def to_dict(self, abridged=False):
        # type: (bool) -> dict[str, dict]

        d = {}
        if abridged == False:
            d["type"] = "ModelPhPropertiesAbridged"
            d["id_num"] = self.id_num
            d["bldg_segments"] = self._get_bldg_segment_dicts()
            d["team"] = self.team.to_dict()
        else:
            d["type"] = "ModelPhProperties"
            d["id_num"] = self.id_num
            d["bldg_segments"] = []
            d["team"] = self.team.to_dict()

        return {"ph": d}

    @classmethod
    def from_dict(cls, _dict, host):
        # type: (dict[str, Any], Any) -> ModelPhProperties
        assert _dict["type"] == "ModelPhProperties", "Expected ModelPhProperties. Got {}.".format(_dict["type"])

        new_prop = cls(host)
        new_prop.id_num = _dict.get("id_num", 0)
        new_prop.team = ProjectTeam.from_dict(_dict.get("team", {}))

        return new_prop

    @staticmethod
    def load_properties_from_dict(data):
        # type: (Dict[str, Dict]) -> Tuple[Dict[str, BldgSegment], ProjectTeam]
        """Load the HB-Model .ph properties from an HB-Model dictionary as Python objects.

        Loaded objects include: BldgSegment, Team, ...

        The function is called when re-serializing an HB-Model object from a
        dictionary. It will load honeybee_ph entities as Python objects and returns
        a tuple of dictionaries with all the de-serialized Honeybee-PH objects.

        Arguments:
        ----------
            data: A dictionary representation of an entire honeybee-core Model.
                 Note that this dictionary must have ModelPhProperties
                in order for this method to successfully apply the .ph properties.

                Note: data is an HB-Model dict and .keys() will include:
                [
                    'display_name', 'identifier', 'tolerance',
                    'angle_tolerance', 'rooms', 'type', 'version',
                    'units', 'orphaned_shades', 'properties'
                ]

        Returns:
        --------
            * tuple[dict[str, BldgSegment], ProjectTeam]: ModelPhProperties Objects
        """
        assert "ph" in data["properties"], "HB-Model Dictionary possesses no ModelPhProperties?"

        bldg_segments_ = {}
        for seg in data["properties"]["ph"]["bldg_segments"]:
            bldg_segments_[seg["identifier"]] = BldgSegment.from_dict(seg)

        team = ProjectTeam.from_dict(data["properties"]["ph"]["team"])

        return bldg_segments_, team

    def apply_properties_from_dict(self, data):
        # type: (Dict[str, Any]) -> None
        """Apply the .ph properties of a dictionary to the host Model of this object.

        This method is called when the HB-Model is de-serialized from a dict back into
        a Python object. In an 'Abridged' HBJSON file, all the property information
        is stored at the model level, not at the sub-model object level. In that case,
        this method is used to apply the right property data back onto all the sub-model
        objects (faces, rooms, apertures, etc).

        Arguments:
        ----------
            * data (dict[str, Any]): A dictionary representation of an entire honeybee-core
                Model. Note that this dictionary must have ModelPhProperties
                in order for this method to successfully apply the .ph properties.

                Note: data is an HB-Model dict and .keys() will include:
                [
                    'display_name', 'identifier', 'tolerance',
                    'angle_tolerance', 'rooms', 'type', 'version',
                    'units', 'orphaned_shades', 'properties'
                ]

        Returns:
        --------
            * None
        """
        assert "ph" in data["properties"], "Dictionary possesses no ModelPhProperties?"

        # collect lists of .ph property dictionaries at the sub-model level (room, face, etc)
        (
            room_ph_dicts,
            face_ph_dicts,
            shd_ph_dicts,
            ap_ph_dicts,
            dr_ph_dicts,
        ) = extensionutil.model_extension_dicts(data, "ph", [], [], [], [], [])

        # re-build all of the .ph property objects from the HB-Model dict as python objects
        bldg_segments, self.team = self.load_properties_from_dict(data)

        # apply the .ph properties to all the sub-model objects in the HB-Model
        for room, room_dict in zip(self.host.rooms, room_ph_dicts):
            if not room_dict:
                continue
            room.properties.ph.apply_properties_from_dict(room_dict, bldg_segments)

        apertures = []
        faces = []
        for hb_room in self.host.rooms:
            for face in hb_room.faces:
                faces.append(face)
                for aperture in face.apertures:
                    apertures.append(aperture)

        for face, face_dict in zip(faces, face_ph_dicts):
            face.properties.ph.apply_properties_from_dict(face_dict)

        for aperture, ap_dict in zip(apertures, ap_ph_dicts):
            aperture.properties.ph.apply_properties_from_dict(ap_dict)



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Utility function to load HB-E-Schedule Objects from a JSON file."""

import json
import os

try:
    from honeybee_energy.schedule.ruleset import ScheduleRuleset
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_energy:\n\t{}".format(e))


def is_schedule(_json_object):
    # type: (dict) -> bool
    """Check if a JSON object is a valid 'ScheduleRuleset' dict."""
    if "type" not in _json_object:
        return False
    if not _json_object["type"] == "ScheduleRuleset":
        return False
    return True


def load_schedules_from_json_file(_schedules_filepath):
    # type: (str) -> dict[str, ScheduleRuleset]
    """Load a set of HBE-ScheduleRuleset object from a JSON file."""
    if not os.path.exists(_schedules_filepath):
        raise ValueError("File not found: {}".format(_schedules_filepath))

    with open(_schedules_filepath, "r") as json_file:
        all_schedules = (ScheduleRuleset.from_dict(d) for d in json.load(json_file) if is_schedule(d))
        return {_.identifier: _ for _ in all_schedules}



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Tools for working with Ladybug.geometry2d.Polygon2D objects."""

import math
from copy import copy

try:
    from typing import List
except ImportError:
    pass  # Python 2.7

try:
    from ladybug_geometry.geometry2d.pointvector import Point2D, Vector2D
    from ladybug_geometry.geometry2d.polygon import Polygon2D
    from ladybug_geometry.geometry3d.face import Face3D
    from ladybug_geometry.geometry3d.plane import Plane
except ImportError as e:
    raise ImportError("Failed to import ladybug_geometry: " + str(e))

try:
    from honeybee_ph_utils import vector3d_tools
except ImportError as e:
    raise ImportError("Failed to import honeybee_ph_utils: " + str(e))


def counterclockwise_angle_between_2_Planes(_plane1, _plane2, _tolerance):
    # type: (Plane, Plane, float) -> float
    """Return the counterclockwise angle (in radians) between two Plane's X-Axes."""

    if not vector3d_tools.vector_equal(_plane1.n, _plane2.n, _tolerance):
        msg = (
            "Error: Cannot calculate the angle between planes with different normal vectors ."
            "Normal vector: {} is not equal to: {}".format(_plane1.n, _plane2.n)
        )
        raise Exception(msg)

    angle_in_radians = vector3d_tools.angle_between_2D_vectors(_plane1.x, _plane2.x)

    # Convert radians to degrees
    angle_degrees = math.degrees(angle_in_radians)

    # Determine the sign of the angle using the cross product
    cross_prod = vector3d_tools.cross_product(_plane1.x, _plane2.x)
    if vector3d_tools.dot_product(cross_prod, _plane1.n) < 0:
        angle_degrees = 360 - angle_degrees

    return math.radians(angle_degrees)


def move_vector_between_two_points(_point1, _point2):
    # type: (Point2D, Point2D) -> Vector2D
    """Return a Vector2D from _point1 to _point2."""

    mv_x = _point1.x - _point2.x
    mv_y = _point1.y - _point2.y
    move_vec = Vector2D(mv_x, mv_y)  # type: ignore
    return move_vec


def translate_polygon2D(_polygon2D, _starting_plane, _target_plane, _tolerance):
    # type: (Polygon2D, Plane, Plane, float) -> Polygon2D
    """Translate (move, rotate) one Polygon2D from its own Plane into another."""

    # ------------------------------------------------------------------------
    # -- Create a Vector2D from the Polygon2D's origin to the _new_plane's
    # -- origin within the new-plane's space.
    target_plane_origin_pt = copy(_target_plane.xyz_to_xy(_target_plane.o))
    polygon2D_origin_pt_in_target_plane_space = copy(_target_plane.xyz_to_xy(_starting_plane.o))

    # ------------------------------------------------------------------------
    # -- Move the starting Polygon2D into the _new_plane's space
    move_vec = move_vector_between_two_points(polygon2D_origin_pt_in_target_plane_space, target_plane_origin_pt)
    moved_polygon2D = _polygon2D.move(move_vec)

    # ------------------------------------------------------------------------
    # -- Rotate the moved Polygon2D to align with the _new_plane's space
    angle_in_radians = counterclockwise_angle_between_2_Planes(
        _target_plane,
        _starting_plane,
        _tolerance,
    )
    rotated_polygon = moved_polygon2D.rotate(angle=angle_in_radians, origin=polygon2D_origin_pt_in_target_plane_space)

    return rotated_polygon


def get_lbt_Face3D_polygon2Ds(_lbt_face3Ds):
    # type: (List[Face3D]) -> List[Polygon2D]
    """Return the LBT-Face3D Polygon2Ds."""

    return [copy(f.polygon2d) for f in _lbt_face3Ds]


def get_lbt_Face3D_planes(_lbt_face3Ds):
    # type: (List[Face3D]) -> List[Plane]
    """Return the LBT-Face3D Planes."""

    return [copy(f.plane) for f in _lbt_face3Ds]


def merge_polygon_2ds(_lbt_polygon_2ds, _tolerance):
    # type: (List[Polygon2D], float) -> List[Polygon2D]
    """Merge together a list of Polygon2Ds."""

    try:
        merged_polygon2Ds = Polygon2D.boolean_union_all(_lbt_polygon_2ds, _tolerance)
    except Exception:
        merged_polygon2Ds = _lbt_polygon_2ds

    return merged_polygon2Ds


def merge_lbt_face_polygons(_lbt_face3Ds, _tolerance):
    # type: (List[Face3D], float) -> List[Polygon2D]
    """Merge together the Polygon2Ds of a list of LBT-Face3Ds."""

    lbt_face3D_polygon2Ds = get_lbt_Face3D_polygon2Ds(_lbt_face3Ds)
    lbt_face3D_planes = get_lbt_Face3D_planes(_lbt_face3Ds)
    reference_plane = _lbt_face3Ds[0].plane

    try:
        # ---------------------------------------------------------------------
        # -- Get all the LBT-Face3D Polygon2Ds in the same Plane-space
        translated_polygon2Ds = []  # type: List[Polygon2D]
        for face3D_poly_2D, face3D_plane in zip(lbt_face3D_polygon2Ds, lbt_face3D_planes):
            translated_polygon2Ds.append(translate_polygon2D(face3D_poly_2D, face3D_plane, reference_plane, _tolerance))

        # ---------------------------------------------------------------------
        # -- Try and merge all the new Polygon2Ds together into a single one.
        merged_polygon2Ds = Polygon2D.boolean_union_all(translated_polygon2Ds, _tolerance)
    except Exception:
        merged_polygon2Ds = list(lbt_face3D_polygon2Ds)

    return merged_polygon2Ds



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""PHI Certification Settings Class."""

from copy import copy

try:
    from typing import Any, Dict, Optional, Type, Union
except ImportError:
    pass  # IronPython 2.7

try:
    from ladybug_geometry.geometry3d.plane import Plane
    from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
except ImportError as e:
    raise ImportError("\nFailed to import ladybug_geometry:\n\t{}".format(e))

try:
    from honeybee_ph import _base
    from honeybee_ph_utils import enumerables
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_ph:\n\t{}".format(e))

# -----------------------------------------------------------------------------


class EnumProperty(object):
    """Descriptor for creating and managing the PHI-Cert Enum items.

    This Descriptor will create a new Enum type when it is called using the
    attribute name and the phpp_version #, This new enum will be used to manage
    the allowable values and clean / filter the user inputs.
    """

    allowed_inputs = {  # type: Dict[str, Dict[int, List]]
        "building_category_type": {
            9: [
                "1-RESIDENTIAL BUILDING",
                "2-NON-RESIDENTIAL BUILDING",
            ],
            10: [],
        },
        "building_use_type": {
            9: [
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "10-DWELLING",
                "11-NURSING HOME / STUDENTS",
                "12-OTHER",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "20-OFFICE / ADMIN. BUILDING",
                "21-SCHOOL",
                "22-OTHER",
            ],
            10: [
                "_",  # PHPP is so stupid. Who did the numbering?... sheesh...
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "10-RESIDENTIAL BUILDING: RESIDENTIAL (DEFAULT)",
                "_",
                "12-RESIDENTIAL BUILDING: OTHER",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "20-NON-RES BUILDING: OFFICE/ADMINISTRATION",
                "21-NON-RES BUILDING: SCHOOL HALF-DAYS (< 7 H)",
                "22-NON-RES BUILDING: SCHOOL FULL-TIME (≥ 7 H)",
                "23-NON-RES.: OTHER",
            ],
        },
        "ihg_type": {
            9: [
                "_",
                "2-STANDARD",
                "3-PHPP CALCULATION ('IHG' WORKSHEET)",
                "4-PHPP CALCULATION ('IHG NON-RES' WORKSHEET)",
            ],
            10: [
                "_",
                "2-STANDARD",
                "3-PHPP-CALCULATION ('IHG' WORKSHEET)",
                "4-PHPP-CALCULATION ('IHG NON-RES' WORKSHEET)",
            ],
        },
        "occupancy_type": {
            9: [
                "1-STANDARD (ONLY FOR RESIDENTIAL BUILDINGS)",
                "2-USER DETERMINED",
            ],
            10: [],
        },
        "certification_type": {
            9: [
                "1-PASSIVE HOUSE",
                "2-ENERPHIT",
                "3-PHI LOW ENERGY BUILDING",
                "4-OTHER",
            ],
            10: [
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "10-PASSIVE HOUSE",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "21-ENERPHIT (COMPONENT METHOD)",
                "22-ENERPHIT (ENERGY DEMAND METHOD)",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "30-PHI LOW ENERGY BUILDING",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "40-OTHER",
            ],
        },
        "certification_class": {
            9: [
                "1-CLASSIC",
                "2-PLUS",
                "3-PREMIUM",
            ],
            10: [
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "10-CLASSIC | PER (RENEWABLE)",
                "11-CLASSIC | PE (NON-RENEWABLE)",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "20-PLUS | PER (RENEWABLE)",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "_",
                "30-PREMIUM | PER (RENEWABLE)",
            ],
        },
        "primary_energy_type": {
            9: [
                "1-PE (NON-RENEWABLE)",
                "2-PER (RENEWABLE)",
            ],
            10: [
                "1-STANDARD",
                "2-PROJECT-SPECIFIC",
            ],
        },
        "enerphit_type": {
            9: [
                "1-COMPONENT METHOD",
                "2-ENERGY DEMAND METHOD",
            ],
            10: [],
        },
        "retrofit_type": {
            9: [
                "1-NEW BUILDING",
                "2-RETROFIT",
                "3-STEP-BY-STEP RETROFIT",
            ],
            10: [
                "1-NEW BUILDING",
                "2-RETROFIT",
                "3-STAGED RETROFIT",
            ],
        },
    }

    def __init__(self, attribute_name, _phpp_version=9):
        # type: (str, int) -> None
        self.attribute_name = attribute_name  # normally the same as the Class Attribute
        self.phpp_version = _phpp_version
        self.enum = self._create_enum_class()
        self.enum.allowed = self.allowed_inputs[self.attribute_name][self.phpp_version]

    def _create_enum_class(self):
        # type: () -> Type[enumerables.CustomEnum]
        """Create a new Enum based on the type of value."""
        return type("{}_phpp_v{}".format(self.attribute_name, self.phpp_version), (enumerables.CustomEnum,), {})

    def __set__(self, instance, value):
        # type: (Any, Optional[Union[str, int]]) -> None
        """Set the enum with the input value on the instance __dict__

        Arguments:
        ----------
            * instance: The descriptor class variable.
            * value: The value to validate and set.
        """

        if value:
            instance.__dict__[self.attribute_name] = self.enum(value)

        if instance.__dict__[self.attribute_name].value == "_":
            msg = "Error: Input value: '{}' for '{}' is not allowed. Check inputs.".format(value, self.attribute_name)
            raise Exception(msg)

    def __get__(self, instance, owner):
        # type: (Any, Any) -> enumerables.CustomEnum
        """
        Arguments:
        ---------
            * instance: The descriptor class variable.
            * owner: The Class which has the descriptor as a class variable.
        Returns:
        --------
            * (enumerables.CustomEnum): The Enum Instance object.
        """
        return instance.__dict__[self.attribute_name]

    def __str__(self):
        return "Validated[{}](storage_name={})".format(self.__class__.__name__, self.attribute_name)


# -----------------------------------------------------------------------------


class _PHPPSettingsBase(object):
    """Base class with methods for use by PHPP-Settings objects."""

    phpp_version = 9  # default
    tfa_override = None  # type: float | None

    def to_dict(self):
        # type: () -> Dict[str, int | str]
        d = {}
        d["phpp_version"] = self.phpp_version
        for k in vars(self).keys():
            attribute = getattr(self, k)
            try:
                d[k] = attribute.value  # -- Try and set the 'value' first
            except AttributeError:
                d[k] = attribute  # -- fallback if that fails for some reason
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, int | str]) -> _PHPPSettingsBase
        new_obj = cls()
        for k in vars(new_obj).keys():
            setattr(new_obj, k, _input_dict[k])
        return new_obj

    def __copy__(self):
        # type: () -> _PHPPSettingsBase
        new_obj = self.__class__()
        for k in vars(new_obj).keys():
            attribute = getattr(self, k)
            try:
                setattr(new_obj, k, attribute.value)  # -- Try and set the 'value' first
            except AttributeError:
                setattr(new_obj, k, attribute)  # -- fallback if that fails for some reason
        return new_obj

    def duplicate(self):
        # type: () -> _PHPPSettingsBase
        return self.__copy__()


class PHPPSettings10(_PHPPSettingsBase):
    """Settings for PHPP v10"""

    phpp_version = 10
    building_use_type = EnumProperty("building_use_type", phpp_version)
    ihg_type = EnumProperty("ihg_type", phpp_version)
    certification_class = EnumProperty("certification_class", phpp_version)
    certification_type = EnumProperty("certification_type", phpp_version)
    primary_energy_type = EnumProperty("primary_energy_type", phpp_version)
    retrofit_type = EnumProperty("retrofit_type", phpp_version)

    def __init__(self):
        # -- Setup the enum defaults
        super(PHPPSettings10, self).__init__()
        self.building_use_type = "10-RESIDENTIAL BUILDING: RESIDENTIAL (DEFAULT)"  # type: EnumProperty
        self.ihg_type = "2-STANDARD"  # type: EnumProperty
        self.certification_class = "10-CLASSIC | PER (RENEWABLE)"  # type: EnumProperty
        self.certification_type = "10-PASSIVE HOUSE"  # type: EnumProperty
        self.primary_energy_type = "1-STANDARD"  # type: EnumProperty
        self.retrofit_type = "1-NEW BUILDING"  # type: EnumProperty
        self.tfa_override = None


class PHPPSettings9(_PHPPSettingsBase):
    """Settings for PHPP v9"""

    phpp_version = 9
    building_category_type = EnumProperty("building_category_type", phpp_version)
    building_use_type = EnumProperty("building_use_type", phpp_version)
    ihg_type = EnumProperty("ihg_type", phpp_version)
    occupancy_type = EnumProperty("occupancy_type", phpp_version)
    certification_type = EnumProperty("certification_type", phpp_version)
    certification_class = EnumProperty("certification_class", phpp_version)
    primary_energy_type = EnumProperty("primary_energy_type", phpp_version)
    enerphit_type = EnumProperty("enerphit_type", phpp_version)
    retrofit_type = EnumProperty("retrofit_type", phpp_version)

    def __init__(self):
        # -- Setup the enum defaults
        super(PHPPSettings9, self).__init__()
        self.building_category_type = "1-RESIDENTIAL BUILDING"  # type: EnumProperty
        self.building_use_type = "10-DWELLING"  # type: EnumProperty
        self.ihg_type = "2-Standard"  # type: EnumProperty
        self.occupancy_type = "1-STANDARD (ONLY FOR RESIDENTIAL BUILDINGS)"  # type: EnumProperty
        self.certification_type = "1-PASSIVE HOUSE"  # type: EnumProperty
        self.certification_class = "1-CLASSIC"  # type: EnumProperty
        self.primary_energy_type = "2-PER (RENEWABLE)"  # type: EnumProperty
        self.enerphit_type = "2-ENERGY DEMAND METHOD"  # type: EnumProperty
        self.retrofit_type = "1-NEW BUILDING"  # type: EnumProperty
        self.tfa_override = None


class PhiCertification(_base._Base):
    """PHI PHPP Certification object with Attributes that vary by version (9 | 10)"""

    def __init__(self, phpp_version=9):
        # type: (int) -> None
        super(PhiCertification, self).__init__()
        self.phpp_version = phpp_version
        if phpp_version == 10:
            self.attributes = PHPPSettings10()
        elif self.phpp_version == 9:
            self.attributes = PHPPSettings9()
        else:
            msg = "Error: Unknown PHPP Version? Got: '{}'".format(self.phpp_version)
            raise Exception(msg)

    def __str__(self):
        return "{}()".format(self.__class__.__name__)

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)

    def to_dict(self):
        # type: () -> Dict
        d = {}
        d["phpp_version"] = self.phpp_version
        d["attributes"] = self.attributes.to_dict()
        d["identifier"] = self.identifier
        d["user_data"] = self.user_data
        d["display_name"] = self.display_name
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict) -> PhiCertification
        new_obj = cls()
        attr_dict = _input_dict["attributes"]
        if attr_dict["phpp_version"] == 10:
            new_obj.attributes = PHPPSettings10.from_dict(attr_dict)
        else:
            new_obj.attributes = PHPPSettings9.from_dict(attr_dict)

        new_obj.identifier = _input_dict.get("identifier", "")
        new_obj.user_data = _input_dict.get("user_data", {})
        new_obj.display_name = _input_dict.get("display_name", "")

        return new_obj

    def __copy__(self):
        # type: () -> PhiCertification
        obj = PhiCertification()
        obj.phpp_version = self.phpp_version
        obj.set_base_attrs_from_source(self)
        obj.attributes = self.attributes.duplicate()
        obj.user_data = copy(self.user_data)

        return obj

    def duplicate(self):
        # type: () -> PhiCertification
        return self.__copy__()

    def move(self, moving_vec3D):
        # type: (Vector3D) -> PhiCertification
        """Move the Phius Certification Object along a vector.

        Args:
            moving_vec3D: A Vector3D with the direction and distance to move the ray.
        """
        new_obj = self.duplicate()
        return new_obj

    def rotate(self, axis_vec3D, angle_degrees, origin_pt3D):
        # type: (Vector3D, float, Point3D) -> PhiCertification
        """Rotate the Phius Certification Object by a certain angle around an axis and origin.

        Right hand rule applies:
        If axis has a positive orientation, rotation will be clockwise.
        If axis has a negative orientation, rotation will be counterclockwise.

        Args:
            axis_vec3D: A Vector3D axis representing the axis of rotation.
            angle_degrees: An angle for rotation in degrees.
            origin_pt3D: A Point3D for the origin around which the object will be rotated.
        """
        new_obj = self.duplicate()
        return new_obj

    def rotate_xy(self, angle_degrees, origin_pt3D):
        # type: (float, Point3D) -> PhiCertification
        """Rotate the Phius Certification Object counterclockwise in the XY plane by a certain angle.

        Args:
            angle_degrees: An angle in degrees.
            origin_pt3D: A Point3D for the origin around which the object will be rotated.
        """
        new_obj = self.duplicate()
        return new_obj

    def reflect(self, plane):
        # type: (Plane) -> PhiCertification
        """Reflected the Phius Certification Object across a plane.

        Args:
            normal_vec3D: A Plane representing the plane across which to reflect.
        """
        new_obj = self.duplicate()
        return new_obj

    def scale(self, scale_factor, origin_pt3D=None):
        # type: (float, Point3D | None) -> PhiCertification
        """Scale the Phius Certification Object by a factor from an origin point.

        Args:
            scale_factor: A number representing how much the line segment should be scaled.
            origin_pt3D: A Point3D representing the origin from which to scale.
                If None, it will be scaled from the World origin (0, 0, 0).
        """
        new_obj = self.duplicate()
        if new_obj.attributes.tfa_override is not None:
            new_obj.attributes.tfa_override = new_obj.attributes.tfa_override * scale_factor
        return new_obj



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-


"""DEPRECATED in favor of the new 'honeybee_phhvac' module."""


from honeybee_energy_ph.hvac import _base


class PhHeatPumpSystem(_base._PhHVACBase):
    message = "PhHeatPumpSystem is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhHeatPumpCoolingParams_Base(_base._PhHVACBase):
    message = "PhHeatPumpCoolingParams_Base is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhHeatPumpCoolingParams_Ventilation(PhHeatPumpCoolingParams_Base):
    message = "PhHeatPumpCoolingParams_Ventilation is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhHeatPumpCoolingParams_Recirculation(PhHeatPumpCoolingParams_Base):
    message = "PhHeatPumpCoolingParams_Recirculation is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhHeatPumpCoolingParams_Dehumidification(PhHeatPumpCoolingParams_Base):
    message = "PhHeatPumpCoolingParams_Dehumidification is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhHeatPumpCoolingParams_Panel(PhHeatPumpCoolingParams_Base):
    message = "PhHeatPumpCoolingParams_Panel is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhHeatPumpCoolingParams:
    message = "PhHeatPumpCoolingParams is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhHeatPumpAnnual(PhHeatPumpSystem):
    message = "PhHeatPumpAnnual is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhHeatPumpRatedMonthly(PhHeatPumpSystem):
    message = "PhHeatPumpRatedMonthly is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhHeatPumpCombined(PhHeatPumpSystem):
    message = "PhHeatPumpCombined is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhHeatPumpSystemBuilder(object):
    message = "PhHeatPumpSystemBuilder is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Utility functions for working with Honeybee-Energy Ventilation Loads and Schedules."""


from honeybee_ph_utils import occupancy


def hb_room_vent_flowrates(_hb_room):
    # type: (room.Room) -> tuple[float, float, float, float]
    """Return the honeybee-Room's four ventilation flow rates in m3/s, or 0 if no ventilation
        program is found on the room.

    Arguments:
    ----------
        * _hb_room (honeybee.room.Room): The honeyebee-Room to get the value for.

    Returns:
    --------
        * tuple[float, float, float, float]:
            - [0] (m3s): The honeybee-Room's ventilation flow_per_person
            - [1] (m3s): The honeybee-Room's ventilation flow_per_area
            - [2] (m3s): The honeybee-Room's ventilation flow_by_air_changes_per_hour
            - [3] (m3s): The honeybee-Room's ventilation flow_per_zone
    """
    default_flow_rates = (0.0, 0.0, 0.0, 0.0)
    try:
        vent_program = _hb_room.properties.energy.ventilation
    except:
        return default_flow_rates

    if vent_program is None:
        # -- Not all programs have a ventilation. If so, return 0 for all flows.
        return default_flow_rates
    else:
        return (
            vent_program.flow_per_person,
            vent_program.flow_per_area,
            vent_program.air_changes_per_hour * _hb_room.volume / 3600,
            vent_program.flow_per_zone,
        )


def hb_room_peak_ventilation_airflow_by_zone(_hb_room):
    # type: (room.Room) -> float
    """Return the Peak Ventilation Airflow (m3/s) for the 'Zone' related elements of a honeybee-Room.

    This will return a sum of the room's flow_per_zone, flow_per_area, and air_changes_per_hour
    but will ignore any occupancy related flow-rates. To get the occupancy-related ventilation
    airflow, use the 'hb_room_peak_ventilation_airflow_by_occupancy' function and to get the
    total airflow (by-zone + by-occupancy) use the 'hb_room_peak_ventilation_airflow_total'

    Arguments:
    ----------
        * _hb_room (room.Room):

    Returns:
    --------
        * float (m/s): Total airflow  (m3/s) For the Ventilation-based airflow (per zone, per area, ach).
    """
    try:
        rm_vent_program = _hb_room.properties.energy.ventilation
        vent_m3s_for_zone = rm_vent_program.flow_per_zone
        vent_m3s_for_area = rm_vent_program.flow_per_area * _hb_room.floor_area
        vent_m3h_for_ach = rm_vent_program.air_changes_per_hour * _hb_room.volume / 3600
        return vent_m3s_for_zone + vent_m3s_for_area + vent_m3h_for_ach
    except AttributeError:
        return 0.0


def hb_room_peak_ventilation_airflow_by_occupancy(_hb_room):
    # type: (room.Room) -> float
    """Return the Peak Ventilation Airflow (m3/s) for the 'Occupancy' related elements of a honeybee-Room.

    This will return the room's flow_per_person, but will ignore any 'zone' related flow-rates.
    such as flow_per_zone or flow_per_area, To get the zone-related ventilation
    airflow, use the 'hb_room_peak_ventilation_airflow_by_zone' function and to get the
    total airflow (by-zone + by-occupancy) use the 'hb_room_peak_ventilation_airflow_total'

    Arguments:
    ----------
        * _hb_room (room.Room):

    Returns:
    --------
        * float (m/s): Total airflow  (m3/s) For the Occupancy-based airflow (per person).
    """

    try:
        peak_occupancy = occupancy.hb_room_peak_occupancy(_hb_room)
        return _hb_room.properties.energy.ventilation.flow_per_person * peak_occupancy
    except AttributeError:
        return 0.0


def hb_room_peak_ventilation_airflow_total(_hb_room):
    # type: (room.Room) -> float
    """Return the total peak ventilation flow rate (m/s). Total of occupancy- and zone-related airlfows.

    Not all honeybee-Rooms have ventilation loads. If that is the case, will return 0.

    Arguments:
    ----------
        * _hb_room (room.Room): The honeyebee room to calculate values for.

    Returns:
    --------
        * float: The total ventilation peak airflow for the honeybee-Room.
    """

    vent_m3s_total = hb_room_peak_ventilation_airflow_by_zone(_hb_room)
    occ_m3s_total = hb_room_peak_ventilation_airflow_by_occupancy(_hb_room)

    return vent_m3s_total + occ_m3s_total



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Called during __init__ to extend the HB-base class 'properties' with a new '_ph_hvac' attribute."""

from honeybee.properties import (
    ApertureProperties,
    DoorProperties,
    FaceProperties,
    ModelProperties,
    RoomProperties,
    ShadeProperties,
)

from .properties.aperture import AperturePhHvacProperties
from .properties.door import DoorPhHvacProperties
from .properties.face import FacePhHvacProperties
from .properties.model import ModelPhHvacProperties
from .properties.room import RoomPhHvacProperties
from .properties.shade import ShadePhHvacProperties

# -----------------------------------------------------------------------------
# Step 1)
# set a private 'ph_hvac' attribute on each relevant HB-Core Property class to None


setattr(ModelProperties, "_ph_hvac", None)
setattr(RoomProperties, "_ph_hvac", None)
setattr(FaceProperties, "_ph_hvac", None)
setattr(ShadeProperties, "_ph_hvac", None)
setattr(ApertureProperties, "_ph_hvac", None)
setattr(DoorProperties, "_ph_hvac", None)


# -----------------------------------------------------------------------------
# Step 2)
# create methods to define the public 'ph_hvac' property instances on each obj.properties container


def model_ph_hvac_properties(self):
    # type: (ModelProperties) -> ModelPhHvacProperties
    if self._ph_hvac is None:
        self._ph_hvac = ModelPhHvacProperties(self.host)
    return self._ph_hvac


def room_ph_hvac_properties(self):
    # type: (RoomProperties) -> RoomPhHvacProperties
    if self._ph_hvac is None:
        self._ph_hvac = RoomPhHvacProperties(self.host)
    return self._ph_hvac


def face_ph_hvac_properties(self):
    # type: (FaceProperties) -> RoomPhHvacProperties
    if self._ph_hvac is None:
        self._ph_hvac = FacePhHvacProperties(self.host)
    return self._ph_hvac


def shade_ph_hvac_properties(self):
    # type: (ShadeProperties) -> RoomPhHvacProperties
    if self._ph_hvac is None:
        self._ph_hvac = ShadePhHvacProperties(self.host)
    return self._ph_hvac


def aperture_ph_hvac_properties(self):
    # type: (ApertureProperties) -> RoomPhHvacProperties
    if self._ph_hvac is None:
        self._ph_hvac = AperturePhHvacProperties(self.host)
    return self._ph_hvac


def door_ph_hvac_properties(self):
    # type: (DoorProperties) -> RoomPhHvacProperties
    if self._ph_hvac is None:
        self._ph_hvac = DoorPhHvacProperties(self.host)
    return self._ph_hvac


# -----------------------------------------------------------------------------
# Step 3)
# add public 'ph_hvac' property methods to the Properties classes


setattr(ModelProperties, "ph_hvac", property(model_ph_hvac_properties))
setattr(RoomProperties, "ph_hvac", property(room_ph_hvac_properties))
setattr(FaceProperties, "ph_hvac", property(face_ph_hvac_properties))
setattr(ShadeProperties, "ph_hvac", property(shade_ph_hvac_properties))
setattr(ApertureProperties, "ph_hvac", property(aperture_ph_hvac_properties))
setattr(DoorProperties, "ph_hvac", property(door_ph_hvac_properties))

# -----------------------------------------------------------------------------
# Step 4_
# Add a public Setter


def model_ph_hvac_properties_setter(self, value):
    # type: (ModelProperties, ModelPhHvacProperties) -> None
    if isinstance(value, ModelPhHvacProperties):
        setattr(self, "_ph_hvac", value)


def room_ph_hvac_properties_setter(self, value):
    # type: (RoomProperties, RoomPhHvacProperties) -> None
    if isinstance(value, RoomPhHvacProperties):
        setattr(self, "_ph_hvac", value)


def face_ph_hvac_properties_setter(self, value):
    # type: (FaceProperties, FacePhHvacProperties) -> None
    if isinstance(value, FacePhHvacProperties):
        setattr(self, "_ph_hvac", value)


def shade_ph_hvac_properties_setter(self, value):
    # type: (ShadeProperties, ShadePhHvacProperties) -> None
    if isinstance(value, ShadePhHvacProperties):
        setattr(self, "_ph_hvac", value)


def aperture_ph_hvac_properties_setter(self, value):
    # type: (ApertureProperties, AperturePhHvacProperties) -> None
    if isinstance(value, AperturePhHvacProperties):
        setattr(self, "_ph_hvac", value)


def door_ph_hvac_properties_setter(self, value):
    # type: (DoorProperties, DoorPhHvacProperties) -> None
    if isinstance(value, DoorPhHvacProperties):
        setattr(self, "_ph_hvac", value)


setattr(ModelProperties, "ph_hvac", ModelProperties.ph_hvac.setter(model_ph_hvac_properties_setter))
setattr(RoomProperties, "ph_hvac", RoomProperties.ph_hvac.setter(room_ph_hvac_properties_setter))
setattr(FaceProperties, "ph_hvac", FaceProperties.ph_hvac.setter(face_ph_hvac_properties_setter))
setattr(ShadeProperties, "ph_hvac", ShadeProperties.ph_hvac.setter(shade_ph_hvac_properties_setter))
setattr(ApertureProperties, "ph_hvac", ApertureProperties.ph_hvac.setter(aperture_ph_hvac_properties_setter))
setattr(DoorProperties, "ph_hvac", DoorProperties.ph_hvac.setter(door_ph_hvac_properties_setter))



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Phius Source Energy Factor Library"""

# -- As per "Phius Core_Phius Zero_Final Modeling Protocol v1.1.pdf"
# -- https://www.phius.org/PHIUS+2021/Phius%20Core_Phius%20Zero_Final%20Modeling%20Protocol%20v1.1.pdf
# -- https://www.phius.org/sites/default/files/2024-07/Phius%20CORE%20and%20ZERO%202024_Final%20Modeling%20Protocol%20v24.1.1.pdf

factors_2018 = {
    "OIL": {"value": 1.1, "unit": "kWh/kWh"},
    "NATURAL_GAS": {"value": 1.1, "unit": "kWh/kWh"},
    "LPG": {"value": 1.1, "unit": "kWh/kWh"},
    "HARD_COAL": {"value": 1.1, "unit": "kWh/kWh"},
    "WOOD": {"value": 0.2, "unit": "kWh/kWh"},
    "ELECTRICITY_MIX": {"value": 2.8, "unit": "kWh/kWh"},
    "ELECTRICITY_PV": {"value": 1.7, "unit": "kWh/kWh"},
    "HARD_COAL_CGS_70_CHP": {"value": 0.8, "unit": "kWh/kWh"},
    "HARD_COAL_CGS_35_CHP": {"value": 1.1, "unit": "kWh/kWh"},
    "HARD_COAL_CGS_0_CHP": {"value": 1.5, "unit": "kWh/kWh"},
    "GAS_CGS_70_CHP": {"value": 0.7, "unit": "kWh/kWh"},
    "GAS_CGS_35_CHP": {"value": 1.1, "unit": "kWh/kWh"},
    "GAS_CGS_0_CHP": {"value": 1.5, "unit": "kWh/kWh"},
    "OIL_CGS_70_CHP": {"value": 0.8, "unit": "kWh/kWh"},
    "OIL_CGS_35_CHP": {"value": 1.1, "unit": "kWh/kWh"},
    "OIL_CGS_0_CHP": {"value": 1.5, "unit": "kWh/kWh"},
}

factors_2021 = {
    "OIL": {"value": 1.1, "unit": "kWh/kWh"},
    "NATURAL_GAS": {"value": 1.1, "unit": "kWh/kWh"},
    "LPG": {"value": 1.1, "unit": "kWh/kWh"},
    "HARD_COAL": {"value": 1.1, "unit": "kWh/kWh"},
    "WOOD": {"value": 0.2, "unit": "kWh/kWh"},
    "ELECTRICITY_MIX": {"value": 1.8, "unit": "kWh/kWh"},
    "ELECTRICITY_PV": {"value": 1.7, "unit": "kWh/kWh"},
    "HARD_COAL_CGS_70_CHP": {"value": 0.8, "unit": "kWh/kWh"},
    "HARD_COAL_CGS_35_CHP": {"value": 1.1, "unit": "kWh/kWh"},
    "HARD_COAL_CGS_0_CHP": {"value": 1.5, "unit": "kWh/kWh"},
    "GAS_CGS_70_CHP": {"value": 0.7, "unit": "kWh/kWh"},
    "GAS_CGS_35_CHP": {"value": 1.1, "unit": "kWh/kWh"},
    "GAS_CGS_0_CHP": {"value": 1.5, "unit": "kWh/kWh"},
    "OIL_CGS_70_CHP": {"value": 0.8, "unit": "kWh/kWh"},
    "OIL_CGS_35_CHP": {"value": 1.1, "unit": "kWh/kWh"},
    "OIL_CGS_0_CHP": {"value": 1.5, "unit": "kWh/kWh"},
}

factors_2024 = {
    "OIL": {"value": 1.1, "unit": "kWh/kWh"},
    "NATURAL_GAS": {"value": 1.1, "unit": "kWh/kWh"},
    "LPG": {"value": 1.1, "unit": "kWh/kWh"},
    "HARD_COAL": {"value": 1.1, "unit": "kWh/kWh"},
    "WOOD": {"value": 0.2, "unit": "kWh/kWh"},
    "ELECTRICITY_MIX": {"value": 2.0, "unit": "kWh/kWh"},
    "ELECTRICITY_PV": {"value": 0.7, "unit": "kWh/kWh"},
    "HARD_COAL_CGS_70_CHP": {"value": 0.8, "unit": "kWh/kWh"},
    "HARD_COAL_CGS_35_CHP": {"value": 1.1, "unit": "kWh/kWh"},
    "HARD_COAL_CGS_0_CHP": {"value": 1.5, "unit": "kWh/kWh"},
    "GAS_CGS_70_CHP": {"value": 0.7, "unit": "kWh/kWh"},
    "GAS_CGS_35_CHP": {"value": 1.1, "unit": "kWh/kWh"},
    "GAS_CGS_0_CHP": {"value": 1.5, "unit": "kWh/kWh"},
    "OIL_CGS_70_CHP": {"value": 0.8, "unit": "kWh/kWh"},
    "OIL_CGS_35_CHP": {"value": 1.1, "unit": "kWh/kWh"},
    "OIL_CGS_0_CHP": {"value": 1.5, "unit": "kWh/kWh"},
}



# -*- Python Version: 2.7 -*-
# -*- coding: utf-8 -*-

"""Classes for calculating Phius Multifamily Elec. Energy as per Phius Multifamily Calculator (v4.2)"""


try:
    from typing import Any, ValuesView
except ImportError:
    pass  # IronPython 2.7

try:
    from honeybee.room import Room
except ImportError as e:
    raise ImportError("Failed to import honeybee: {}".format(e))

try:
    from honeybee_energy.load import equipment
    from honeybee_energy.properties.room import RoomEnergyProperties
    from honeybee_energy.schedule import ruleset
except ImportError as e:
    raise ImportError("Failed to import honeybee_energy: {}".format(e))

try:
    from honeybee_energy_ph.load import phius_residential
except ImportError as e:
    raise ImportError("Failed to import honeybee_energy_ph: {}".format(e))

try:
    from honeybee_ph import space
except ImportError as e:
    raise ImportError("Failed to import honeybee_ph: {}".format(e))

try:
    from ph_units.converter import convert
except ImportError as e:
    raise ImportError("Failed to import ph_units: {}".format(e))


class PhiusResidentialStory(object):
    """Represents one Residential Story of the Phius Multifamily Calculator (v4.2)"""

    LIGHTING_INT_HE_FRAC = 1.0
    LIGHTING_EXT_HE_FRAC = 1.0
    LIGHTING_GARAGE_HE_FRAC = 1.0

    def __init__(self, _hb_rooms, _area_unit):
        # type: (list[Room], str) -> None

        try:
            self.story_number = _hb_rooms[0].story
        except:
            self.story_number = 1

        self.total_floor_area_ft2 = self.calc_story_floor_area_ft2(_hb_rooms, _area_unit)
        self.total_number_dwellings = self.calc_num_dwellings(_hb_rooms)
        self.total_number_bedrooms = self.calc_story_bedrooms(_hb_rooms)

        self.design_occupancy = self.calc_passive_house_occupancy(_hb_rooms)
        self.mel = phius_residential.misc_electrical(
            self.total_number_bedrooms, self.total_floor_area_ft2, self.total_number_dwellings
        )
        self.lighting_int = phius_residential.lighting_interior(
            self.total_floor_area_ft2, self.LIGHTING_INT_HE_FRAC, self.total_number_dwellings
        )
        self.lighting_ext = phius_residential.lighting_exterior(
            self.total_floor_area_ft2, self.LIGHTING_EXT_HE_FRAC, self.total_number_dwellings
        )
        self.lighting_garage = phius_residential.lighting_garage(
            self.LIGHTING_GARAGE_HE_FRAC, self.total_number_dwellings
        )

    @property
    def story_number(self):
        # type: () -> str
        """Return the story number."""
        return str(self._story_number)

    @story_number.setter
    def story_number(self, value):
        # type: (Any) -> None
        """Set the story number."""
        # -- If the input is a single integer value, enforce padding
        try:
            self._story_number = "{:03d}".format(int(value))
        except ValueError:
            self._story_number = str(value)

    def calc_story_floor_area_ft2(self, _hb_rooms, _area_unit):
        # type: (list[Room], str) -> float
        area = sum(space.weighted_floor_area for rm in _hb_rooms for space in getattr(rm.properties, "ph").spaces)
        area_ft2 = convert(area, _area_unit, "FT2")
        if area_ft2 is None:
            raise ValueError("Error: Failed to convert '{}' floor area from {} to FT2".format(area, _area_unit))
        return area_ft2

    def calc_story_bedrooms(self, _hb_rooms):
        # type: (list[Room]) -> int
        return sum(getattr(rm.properties, "energy").people.properties.ph.number_bedrooms for rm in _hb_rooms)

    def calc_passive_house_occupancy(self, _hb_rooms):
        # type: (list[Room]) -> float
        return sum(getattr(rm.properties, "energy").people.properties.ph.number_people for rm in _hb_rooms)

    def calc_num_dwellings(self, _hb_rooms):
        # type: (list[Room]) -> int
        ph_dwelling_objs = {r.properties.energy.people.properties.ph.dwellings for r in _hb_rooms}  # type: ignore
        return sum(d.num_dwellings for d in ph_dwelling_objs)

    def __lt__(self, other):
        # type: (PhiusResidentialStory) -> bool
        return self.story_number < other.story_number

    def __eq__(self, other):
        # type: (PhiusResidentialStory) -> bool
        return self.story_number == other.story_number

    def __str__(self):
        return "{}(total_floor_area_ft2={}, total_number_dwellings={}, total_number_bedrooms={})".format(
            self.__class__.__name__,
            self.total_floor_area_ft2,
            self.total_number_dwellings,
            self.total_number_bedrooms,
        )

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)


# -----------------------------------------------------------------------------
# -- Non-Res


class PhiusNonResProgram(object):
    """An individual Phius Non-Res Program Type."""

    def __init__(self):
        self.name = "__unnamed_nonres_program__"
        self.usage_days_yr = 365
        self.operating_hours_day = 0.0
        self.lighting_W_per_m2 = 0.0
        self.mel_kWh_m2_yr = 0.0
        self.mel_w_m2 = 0.0

    @property
    def lighting_W_per_ft2(self):
        # type: () -> float
        return self.lighting_W_per_m2 * 0.09290304

    @property
    def mel_kWh_ft2_yr(self):
        # type: () -> float
        return self.mel_kWh_m2_yr * 0.09290304

    def to_phius_mf_workbook(self):
        # type: () -> str
        """Returns a text block formatted to match the Phius MF Calculator."""
        return ",".join(
            [
                str(self.name),
                str(self.usage_days_yr),
                str(self.operating_hours_day),
                str(self.lighting_W_per_ft2),
                str(self.mel_kWh_ft2_yr),
            ]
        )

    @classmethod
    def from_hb_room(cls, _hb_room):
        # type: (Room) -> PhiusNonResProgram
        """Returns a new PhiusNonResProgram object with attributes based on an HBE-Room."""
        obj = cls()

        # -- type Alias
        rm_prop_energy = getattr(_hb_room.properties, "energy")  # type: RoomEnergyProperties

        obj.name = rm_prop_energy.program_type.display_name
        if not isinstance(rm_prop_energy.lighting.schedule, ruleset.ScheduleRuleset):
            raise ValueError(
                "Error: Room Lighting Schedule must be a ScheduleRuleset object."
                "The Lighting Schedule {} from room {} is of type {} which is not compatible wth Honeybee-PH".format(
                    rm_prop_energy.lighting.schedule.display_name,
                    _hb_room.display_name,
                    type(rm_prop_energy.lighting.schedule),
                )
            )
        obj.operating_hours_day = obj._operating_hours_day_from_hb_schedule(rm_prop_energy.lighting.schedule)
        obj.lighting_W_per_m2 = rm_prop_energy.lighting.watts_per_area
        obj.mel_kWh_m2_yr = obj._calc_mel_kWh_per_m2_from_hb_elec(rm_prop_energy.electric_equipment)
        obj.mel_w_m2 = rm_prop_energy.electric_equipment.watts_per_area

        return obj

    def _calc_mel_kWh_per_m2_from_hb_elec(self, _hb_elec_equip):
        # type: (equipment.ElectricEquipment) -> float
        """Returns the total Elec. Equip kWh for an HBE-Electric-Equipment object."""

        # -- Calc annual full-load hours
        if not isinstance(_hb_elec_equip.schedule, ruleset.ScheduleRuleset):
            raise ValueError(
                "Error: Electric Equipment Schedule must be a ScheduleRuleset object."
                "The Electric Equipment Schedule {} is of type {} which is not compatible wth Honeybee-PH".format(
                    _hb_elec_equip.schedule.display_name, type(_hb_elec_equip.schedule)
                )
            )
        annual_full_load_hours = sum(_hb_elec_equip.schedule.values())

        # -- Calc total usage in kWh
        return (_hb_elec_equip.watts_per_area * annual_full_load_hours) / 1000

    def _operating_hours_day_from_hb_schedule(self, _hb_lght_sched):
        # type: (ruleset.ScheduleRuleset) -> float
        """Return a daily operating period (num. hours) from an HB-Lighting-Schedule"""
        operating_frac = sum(_hb_lght_sched.values()) / len(_hb_lght_sched.values())
        return operating_frac * 24

    def __str__(self):
        return "{}(name={})".format(self.__class__.__name__, self.name)

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)


class PhiusNonResProgramCollection(object):
    """Collection of Phius Non-Res Program Types."""

    def __init__(self):
        self._collection = {}

    def add_program(self, _program, _key=None):
        # type: (PhiusNonResProgram, str | None) -> None
        """Adds a new PhiusNonResProgram to the collection."""
        if not _key:
            _key = _program.name
        self._collection[_key] = _program

    @property
    def programs(self):
        # type: () -> ValuesView[PhiusNonResProgram]
        """Returns a ValuesView of the PhiusNonResPrograms in the collection dict."""
        return self._collection.values()

    def __getitem__(self, key):
        # type: (str) -> PhiusNonResProgram
        return self._collection[key]

    def to_phius_mf_workbook(self):
        # type: () -> list[str]
        """Returns a text block formatted to match the Phius MF Calculator."""
        return [prog.to_phius_mf_workbook() for prog in self.programs]


class PhiusNonResRoom(object):
    """A single Phius Non-Res Space."""

    def __init__(self):
        self.multiplier = 1
        self.occupancy_sensor = "N"
        self.name = "_unnamed_phius_nonres_room"
        self.reference_floor_area_m2 = 0.0
        self.misc_mel = 0.0  # kWh/yr
        self.program_type = PhiusNonResProgram()

    @property
    def reference_floor_area_ft2(self):
        # type: () -> float
        return self.reference_floor_area_m2 * 10.7639

    @property
    def mel_kWh_yr(self):
        # type: () -> float
        """Total yearly MEL kWh EXCLUDING the Misc MEL"""
        return self.reference_floor_area_m2 * self.program_type.mel_kWh_m2_yr

    @property
    def mel_w_m2(self):
        # type: () -> float
        return self.program_type.mel_w_m2

    @property
    def total_mel_kWh(self):
        """Total yearly MEL kWh INCLUDING the Misc MEL"""
        # type: () -> float
        return self.mel_kWh_yr + self.misc_mel

    @property
    def total_lighting_kWh(self):
        # type: () -> float
        return (
            self.program_type.usage_days_yr
            * self.program_type.operating_hours_day
            * self.program_type.lighting_W_per_m2
            * self.reference_floor_area_m2
        ) / 1000

    def to_phius_mf_workbook(self):
        # type: () -> str
        """Returns a string representation that matches the Phius MF Calculator."""
        return ",".join(
            [
                str(self.program_type.name),
                "",  # Blank
                "",  # Number
                str(self.name),
                str(self.occupancy_sensor),
                str(self.multiplier),
                str(self.reference_floor_area_ft2),
                str(self.mel_w_m2),
            ]
        )

    def to_phius_mf_workbook_results(self):
        return ",".join(
            [
                str(self.program_type.lighting_W_per_ft2),
                str(self.program_type.usage_days_yr),
                str(self.program_type.operating_hours_day),
                str(self.program_type.mel_kWh_ft2_yr),
                str(self.total_lighting_kWh),
                str(self.mel_kWh_yr),
            ]
        )

    @classmethod
    def from_ph_space(cls, _ph_space, _area_unit):
        # type: (space.Space, str) -> PhiusNonResRoom
        """Returns a new PhiusNonResSpace with properties based on a PH-Space."""
        obj = cls()

        obj.name = _ph_space.full_name
        obj.reference_floor_area_m2 = convert(_ph_space.weighted_net_floor_area, _area_unit, "M2") or 0.0
        if _ph_space.host is not None:
            obj.program_type = PhiusNonResProgram.from_hb_room(_ph_space.host)

        return obj

    def __str__(self):
        return "{}(name={}, program_type={}," "misc_mel={}, ref_floor_area={}m2)".format(
            self.__class__.__name__,
            self.name,
            self.program_type,
            self.misc_mel,
            self.reference_floor_area_m2,
        )

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Utility functions for working with Honeybee-Energy Occupancy Loads and Schedules"""

import statistics


def hb_room_ppl_per_area(_hb_room):
    # type: (room.Room) -> float
    """Returns a honeybee-Room's occupancy load (people_per_area).

    Note all  honeybee-Rooms have an occupancy (stairs, etc) and so if there is
    no energy.pepple found, will return 0

    Arguments:
    ----------
        * _hb_room (honeybee.room.Room): The honeyebee-Room to get the value for.

    Returns:
    --------
        * (float): The honeybee-Room's occupancy load (people_per_area)
    """
    try:
        return _hb_room.properties.energy.people.people_per_area
    except AttributeError:
        return 0.0


def hb_room_peak_occupancy(_hb_room):
    # type: (room.Room) ->  float
    """Returns a peak occupancy (# ppl) of a honeybee-Room.

    Not all honeybee rooms have an occupancy (stairs, etc..) and so if there is
    no energy.people found, will return 0.

    Arguments:
    ----------
        * _hb_room (honeybee.room.Room): The honeyebee-Room to calculate values for.

    Returns:
    --------
        * (float): The honeybee-Room's peak occupancy (total number of people)
    """
    try:
        ppl_per_area = _hb_room.properties.energy.people.people_per_area
    except AttributeError:
        ppl_per_area = 0

    return _hb_room.floor_area * ppl_per_area


def hb_room_annual_avg_occupancy(_hb_room):
    # type: (room.Room) -> float
    """Returns the annual average occupancy (# ppl) of a honeybee-Room.

    Will return the 'mean_occupancy' if there is one on the Schedule. Otherwise
    will calculate the value.

    Arguments:
    ----------
        * _hb_room (room.Room): The honeyebee-Room to calculate values for.

    Returns:
    --------
        * (float): The honeybee-Room's average annual occupancy (total number of people).
    """

    peak_occupancy = hb_room_peak_occupancy(_hb_room)
    try:
        mean_occupancy = _hb_room.properties.energy.people.occupancy_schedule.mean_occupancy
    except AttributeError:
        mean_occupancy = statistics.mean(_hb_room.properties.energy.people.occupancy_schedule.values())

    return peak_occupancy * mean_occupancy



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Honeybee-PH-HVAC-Equipment: Fuels Enumeration"""

# -- Boiler Fuels
NATURAL_GAS = "NATURAL_GAS"
OIL = "OIL"
WOOD_LOG = "WOOD_LOG"
WOOD_PELLET = "WOOD_PELLET"
ELECTRICITY = "ELECTRICITY"

# -- District Energy Carriers
HARD_COAL_CGS_70_PHC = "HARD_COAL_CGS_70_PHC"
HARD_COAL_CGS_35_PHC = "HARD_COAL_CGS_35_PHC"
HARD_COAL_HS_0_PHC = "HARD_COAL_HS_0_PHC"
GAS_CGS_70_PHC = "GAS_CGS_70_PHC"
GAS_CGS_35_PHC = "GAS_CGS_35_PHC"
GAS_HS_0_PHC = "GAS_HS_0_PHC"
OIL_CGS_70_PHC = "OIL_CGS_70_PHC"
OIL_CGS_35_PHC = "OIL_CGS_35_PHC"
OIL_HS_0_PHC = "OIL_HS_0_PHC"



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""HB-Face Passive House (PH) Properties."""
try:
    from typing import Any, Dict
except ImportError:
    pass  # Python 2.7


class FacePhProperties(object):
    def __init__(self, _host):
        self._host = _host
        self.id_num = 0

    @property
    def host(self):
        return self._host

    def duplicate(self, new_host=None):
        # type: (Any) -> FacePhProperties
        _host = new_host or self._host
        new_properties_obj = FacePhProperties(_host)
        new_properties_obj.id_num = self.id_num

        return new_properties_obj

    def ToString(self):
        return self.__repr__()

    def __repr__(self):
        return "Face Passive House Properties: [host: {}]".format(self.host.display_name)

    def to_dict(self, abridged=False):
        # type: (bool) -> dict[str, dict]
        d = {}
        t = "FacePhProperties" if not abridged else "FacePhPropertiesAbridged"
        d.update({"type": t})
        d.update({"id_num": self.id_num})

        return {"ph": d}

    @classmethod
    def from_dict(cls, data, host):
        # type: (dict, Any) -> FacePhProperties
        assert data["type"] == "FacePhProperties", "Expected FacePhProperties. Got {}.".format(data["type"])

        new_prop = cls(host)
        new_prop.id_num = data.get("id_num", 0)

        return new_prop

    def apply_properties_from_dict(self, _face_prop_dict):
        # type: (Dict[str, Any]) -> None
        """Apply properties from an FacePhPropertiesAbridged dictionary.

        Arguments:
        ----------
            * _face_prop_dict (dict): A FacePhPropertiesAbridged dictionary loaded from
                the Face object itself. Unabridged.

        Returns:
        --------
            * None
        """
        return None



# -*- Python Version: 2.7 -*-
# -*- coding: utf-8 -*-

"""Functions for Automating some common Sky Matrix operations."""

import math
import os
import subprocess

try:
    from typing import Any, Tuple
except ImportError:
    pass  # IronPython

try:
    import ladybug.analysisperiod as ap
    import ladybug.epw as epw
    from ladybug.config import folders as lb_folders
    from ladybug.dt import DateTime
    from ladybug.viewsphere import view_sphere
    from ladybug.wea import Wea
except ImportError as e:
    raise ImportError("\nFailed to import ladybug:\n\t{}".format(e))

try:
    from ladybug_geometry.geometry2d.pointvector import Vector2D
except ImportError as e:
    raise ImportError("\nFailed to import ladybug_geometry:\n\t{}".format(e))

try:
    from ladybug_rhino.grasshopper import objectify_output
    from ladybug_rhino.togeometry import to_vector2d
except ImportError as e:
    raise ImportError("\nFailed to import ladybug_rhino:\n\t{}".format(e))

try:
    from honeybee_radiance.config import folders as hb_folders
    from lbt_recipes.version import check_radiance_date
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_radiance:\n\t{}".format(e))


from honeybee_ph_utils.input_tools import memoize

# -----------------------------------------------------------------------------
# constants for converting RGB values output by gendaymtx to broadband radiation
PATCHES_PER_ROW = {
    1: view_sphere.TREGENZA_PATCHES_PER_ROW + (1,),
    2: view_sphere.REINHART_PATCHES_PER_ROW + (1,),
}
PATCH_ROW_COEFF = {
    1: view_sphere.TREGENZA_COEFFICIENTS,
    2: view_sphere.REINHART_COEFFICIENTS,
}


@memoize
def create_analysis_period_hoys(_period):
    # type: (Tuple[int, int]) -> Tuple
    """Return a new Ladybug Analysis Period HOYS."""

    anp = ap.AnalysisPeriod(
        st_month=_period[0],
        st_day=1,
        st_hour=0,
        end_month=_period[1],
        end_day=31,
        end_hour=23,
        timestep=1,
        is_leap_year=False,
    )
    return anp.hoys


def broadband_radiation(patch_row_str, row_number, wea_duration, sky_density=1):
    """Parse a row of gendaymtx RGB patch data in W/sr/m2 to radiation in kWh/m2.

    This includes applying broadband weighting to the RGB bands, multiplication
    by the steradians of each patch, and multiplying by the duration of time that
    they sky matrix represents in hours.

    Args:
        patch_row_str: Text string for a single row of RGB patch data.
        row_number: Integer for the row number that the patch corresponds to.
        sky_density: Integer (either 1 or 2) for the density.
        wea_duration: Number for the duration of the Wea in hours. This is used
            to convert between the average value output by the command and the
            cumulative value that is needed for all ladybug analyses.
    """
    R, G, B = patch_row_str.split(" ")
    weight_val = 0.265074126 * float(R) + 0.670114631 * float(G) + 0.064811243 * float(B)
    return weight_val * PATCH_ROW_COEFF[sky_density][row_number] * wea_duration / 1000


def parse_mtx_data(data_str, wea_duration, sky_density=1):
    """Parse a string of Radiance gendaymtx data to a list of radiation-per-patch.

    This function handles the removing of the header and the conversion of the
    RGB irradiance-=per-steraidian values to broadband radiation. It also removes
    the first patch, which is the ground and is not used by Ladybug.

    Args:
        data_str: The string that has been output by gendaymtx to stdout.
        wea_duration: Number for the duration of the Wea in hours. This is used
            to convert between the average value output by the command and the
            cumulative value that is needed for all ladybug analyses.
        sky_density: Integer (either 1 or 2) for the density.
    """
    # split lines and remove the header, ground patch and last line break
    data_lines = data_str.split("\n")
    patch_lines = data_lines[9:-1]

    # loop through the rows and convert the radiation RGB values
    broadband_irr = []
    patch_counter = 0
    for i, row_patch_count in enumerate(PATCHES_PER_ROW[sky_density]):
        row_slice = patch_lines[patch_counter : patch_counter + row_patch_count]
        irr_vals = (broadband_radiation(row, i, wea_duration, sky_density) for row in row_slice)
        broadband_irr.extend(irr_vals)
        patch_counter += row_patch_count
    return broadband_irr


@memoize
def gen_matrix(_epw_file, _period, _north):
    # type: (str, Tuple[int, int], Any) -> Any

    epw_data = epw.EPW(_epw_file)
    _location = epw_data.location
    high_density_ = None
    _ground_ref_ = None
    _hoys_ = create_analysis_period_hoys(_period)
    _direct_rad = epw_data.direct_normal_radiation
    _diffuse_rad = epw_data.diffuse_horizontal_radiation
    _folder_ = None

    # -------------------------------------------------------------------------
    # check the installed Radiance date and get the path to the gendaymtx executable
    check_radiance_date()
    gendaymtx_exe = (
        os.path.join(hb_folders.radbin_path, "gendaymtx.exe")
        if os.name == "nt"
        else os.path.join(hb_folders.radbin_path, "gendaymtx")
    )

    # -------------------------------------------------------------------------
    # process and set defaults for all of the global inputs
    if _north is not None:  # process the north_
        try:
            north_ = math.degrees(to_vector2d(_north).angle_clockwise(Vector2D(0, 1)))
        except AttributeError:  # north angle instead of vector
            north_ = float(_north)
    else:
        north_ = 0
    density = 2 if high_density_ else 1
    ground_r = 0.2 if _ground_ref_ is None else _ground_ref_

    # -------------------------------------------------------------------------
    # filter the radiation by _hoys if they are input
    if len(_hoys_) != 0:
        _direct_rad = _direct_rad.filter_by_hoys(_hoys_)
        _diffuse_rad = _diffuse_rad.filter_by_hoys(_hoys_)

    # -------------------------------------------------------------------------
    # create the wea and write it to the default_epw_folder
    wea = Wea(_location, _direct_rad, _diffuse_rad)
    wea_duration = len(wea) / wea.timestep
    wea_folder = _folder_ if _folder_ is not None else os.path.join(lb_folders.default_epw_folder, "sky_matrices")
    metd = _direct_rad.header.metadata
    wea_basename = metd["city"].replace(" ", "_") if "city" in metd else "unnamed"
    wea_path = os.path.join(wea_folder, wea_basename)
    wea_file = wea.write(wea_path)

    # -------------------------------------------------------------------------
    # execute the Radiance gendaymtx command
    use_shell = True if os.name == "nt" else False

    # -------------------------------------------------------------------------
    # command for direct patches
    cmds = [gendaymtx_exe, "-m", str(density), "-d", "-O1", "-A", wea_file]
    process = subprocess.Popen(cmds, stdout=subprocess.PIPE, shell=use_shell)
    stdout = process.communicate()
    dir_data_str = stdout[0]

    # -------------------------------------------------------------------------
    # command for diffuse patches
    cmds = [gendaymtx_exe, "-m", str(density), "-s", "-O1", "-A", wea_file]
    process = subprocess.Popen(cmds, stdout=subprocess.PIPE, shell=use_shell)
    stdout = process.communicate()
    diff_data_str = stdout[0]

    # parse the data into a single matrix
    dir_vals = parse_mtx_data(dir_data_str, wea_duration, density)
    diff_vals = parse_mtx_data(diff_data_str, wea_duration, density)

    # -------------------------------------------------------------------------
    # collect sky metadata like the north, which will be used by other components
    metadata = [north_, ground_r]
    if _hoys_:
        metadata.extend([DateTime.from_hoy(h) for h in (_hoys_[0], _hoys_[-1])])
    else:
        metadata.extend([wea.analysis_period.st_time, wea.analysis_period.end_time])
    for key, val in _direct_rad.header.metadata.items():
        metadata.append("{} : {}".format(key, val))

    # -------------------------------------------------------------------------
    # wrap everything together into an object to output from the component
    mtx_data = (metadata, dir_vals, diff_vals)
    sky_mtx = objectify_output("Cumulative Sky Matrix", mtx_data)

    return sky_mtx



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Preview Python Object and sub-objects"""

import uuid


def object_preview(_obj, _full=False, _level=1):
    # type: (object, bool, int) ->  None
    """ "Recursively print all of an object's attributes and child-objects (and attributes).

    Arguments:
    ----------
        * _obj (object): A python object to print the attributes of.
        * _full (bool): default=False. True=Print all the 'backend' attributes like 'identifier'.
            False=ignore the 'backend' attributes.
        * _level (int): default=1. The 'inset' level from the left.

    Return:
    -------
        * None

    """

    if not _obj:
        return None

    if not hasattr(_obj, "__dict__"):
        print("{} object has no __dict__ attribute?".format(_obj))
        return None

    print("{}CLASS:: {}".format(" " * _level, _obj.__class__.__name__))

    for k, v in _obj.__dict__.items():
        if not _full:
            # Skip over some of the basic Honeybee-PH back-end attributes
            if type(v) == type(uuid.uuid4()):
                continue
            if "identifier" in k:
                continue
            if "user_data" == k:
                continue
            if "_host" == k:
                continue

        if hasattr(v, "__dict__"):
            print("{}+ [ {} ] ::: {}".format(" " * _level, k, v))

            # Recursively step through all the child objects
            object_preview(v, _full, _level=_level + 8)
        else:
            print("{}> {} ::: {}".format(" " * _level, k, v))



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""A simplified custom Enum class since IronPython doesn't have an enum."""

try:
    from typing import Any, Union
except ImportError:
    pass  # IronPython


class ValueNotAllowedError(Exception):
    def __init__(self, _in, _enum):
        self.message = "Value: {} not allowed for enum {}.\nValid input: {} ".format(
            str(_in), _enum.__class__.__name__, _enum.allowed
        )
        super(ValueNotAllowedError, self).__init__(self.message)


class CustomEnum(object):
    """A simplified custom Enum class since IronPython doesn't have an enum.

    Example usage:
    ```python
    class PhFoundationType(enumerables.CustomEnum):
        allowed = [
            "1-HEATED_BASEMENT",
            "2-UNHEATED_BASEMENT",
            "3-SLAB_ON_GRADE",
            "4-VENTED_CRAWLSPACE",
            "5-NONE",
        ]

        def __init__(self, _value=1):
            # type: (Union[str, int]) -> None
            super(PhFoundationType, self).__init__(_value)

    new_foundation_type = PhFoundationType("1-HEATED_BASEMENT")
    ```
    """

    allowed = []  # type: list[str]

    def __init__(self, _value="", _index_offset=-1):
        # type: (Union[str, int], int) -> None

        # Standard offset is 1 since most listings start at 1 (but not all)
        self.index_offset = _index_offset

        self._value = ""
        self.value = _value

    @property
    def allowed_upper(self):
        # type: () -> list[str]
        return [_.upper() for _ in self.allowed]

    @property
    def value(self):
        # type: () -> str
        return str(self._value)

    @value.setter
    def value(self, _in):
        # type: (Union[str, int]) -> None
        """Allows the user to set the .value as one of the allowed values. If an
        integer is passed in, will attempt to find the corresponding value from the
        allowed-values list (1-based, ie: user-input '1' -> self.allowed.index(0) ).
        """

        if str(_in).upper() in self.allowed_upper:
            self._value = str(_in).upper()
        else:
            try:
                input = int(_in) + self.index_offset
                self._value = self.allowed_upper[input]
            except:
                raise ValueNotAllowedError(_in, self)

    @property
    def number(self):
        # type: () -> int
        """Returns the index pos of self.value (usually 1-based)"""
        return self.allowed_upper.index(self.value) - self.index_offset

    def __str__(self):
        return "{}(_value={} [number={}])".format(self.__class__.__name__, self.value, self.number)

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)

    def to_dict(self):
        # type: () -> dict[str, Any]
        d = {}
        d["value"] = self.value
        return d

    @classmethod
    def from_dict(cls, _dict):
        # type: (dict[str, Any]) -> CustomEnum
        obj = cls(_dict.get("value", 1))
        return obj

    def __eq__(self, other):
        # type: (CustomEnum) -> bool
        return self.value == other.value and self.__class__ == other.__class__

    def __hash__(self):
        # type: () -> int
        return hash(self.value) + hash(self.__class__)



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Package: Honeybee-PH-HVAC

These classes are designed to be used within a Honeybee-editing environment, such as Rhino or
Grasshopper. These classes will allow for Passive House specific HVAC data to be added to certain
relevant Honeybee entities and logged for later use.
"""

# load all functions that extends honeybee core library
# Make sure the load `honeybee_energy` before anything else to avoid import errors.
import honeybee_energy._extend_honeybee

import honeybee_phhvac._extend_honeybee_ph_hvac



# -*- Python Version: 2.7 -*-
# -*- coding: utf-8 -*-

"""HB-PH Electric Equipment Types."""

try:
    from honeybee_ph_utils import enumerables
except ImportError as e:
    raise ImportError("Failed to import honeybee_ph_utils: {}".format(e))


# -----------------------------------------------------------------------------
# - Type Enums


class PhDishwasherType(enumerables.CustomEnum):
    allowed = [
        "1-DHW CONNECTION",
        "2-COLD WATER CONNECTION",
    ]

    def __init__(self, _value=1):
        # type: (int | str) -> None
        super(PhDishwasherType, self).__init__(_value)


class PhClothesWasherType(enumerables.CustomEnum):
    allowed = [
        "1-DHW CONNECTION",
        "2-COLD WATER CONNECTION",
    ]

    def __init__(self, _value=1):
        # type: (int | str) -> None
        super(PhClothesWasherType, self).__init__(_value)


class PhClothesDryerType(enumerables.CustomEnum):
    allowed = [
        "1-CLOTHES LINE",
        "2-DRYING CLOSET (COLD!)",
        "3-DRYING CLOSET (COLD!) IN EXTRACT AIR",
        "4-CONDENSATION DRYER",
        "5-ELECTRIC EXHAUST AIR DRYER",
        "6-GAS EXHAUST AIR DRYER",
    ]

    def __init__(self, _value=1):
        # type: (int | str) -> None
        super(PhClothesDryerType, self).__init__(_value)


class PhCookingType(enumerables.CustomEnum):
    allowed = [
        "1-ELECTRICITY",
        "2-NATURAL GAS",
        "3-LPG",
    ]

    def __init__(self, _value=1):
        # type: (int | str) -> None
        super(PhCookingType, self).__init__(_value)
